/* File:      locations.C
   Date:      6/21/02  (written fall 2001)
   Author:    C. Paciorek
   Purpose:   code for dealing with locational information
   Compile:   g++ -c
   Usage:     
   Arguments: 
   Requires:  base.h,locations.h,projection.h,permtools.h
   Reference: George & Liu 1981: Computer solution of large sparse positive definite systems - this source has the algorithms and storage format of the adjacency info and the shuffling function
*/


#include "base.h"
#include "projection.h"
#include "locations.h"
#include "permtools.h"

locn_info::locn_info(string lat_in_fn,string lon_in_fn,string kernel_lat_in_fn,string kernel_lon_in_fn,string perm_fn, string invp_fn,int format_in,double range_in,int num_latpoints_in,int num_lonpoints_in,int num_kernels_in){
  int i,j;

  if(format_in==NON_GRID){
    cerr << "Error: code for non-grid locations not yet written\n";
  }
  else{
    range=range_in;                          // maximal non-zero correlation
    format=format_in;
    num_lonpoints=num_lonpoints_in;
    num_latpoints=num_latpoints_in;
    num_gridpoints=num_lonpoints*num_latpoints;
    num_kernels=num_kernels_in;

    // data and kernel lats and lons
    mat_pp<double> tmplat(1,num_latpoints);
    mat_pp<double> tmplon(1,num_lonpoints);
    tmplat.read(lat_in_fn,BY_ROW,RAGGED);    // reads in the latitudes
    tmplon.read(lon_in_fn,BY_ROW,RAGGED);    // reads in the longitudes
    tmplat.getvec(0,ROW,&lat_grid);           
    tmplon.getvec(0,ROW,&lon_grid);           
    // this change from "lat_grid=tmplat.getvec(0,ROW)" made because of warning about tmp being used uninstantiated, although the results of runs seem same regardless of which version of getvec I use
    
    if(num_kernels){
      mat_pp<double> tmpklat(1,num_kernels);
      mat_pp<double> tmpklon(1,num_kernels);
      tmpklat.read(kernel_lat_in_fn,BY_ROW,RAGGED);    // reads in the latitudes
      tmpklon.read(kernel_lon_in_fn,BY_ROW,RAGGED);    // reads in the longitudes
      tmpklat.getvec(0,ROW,&kernel_lat);           
      tmpklon.getvec(0,ROW,&kernel_lon);           
    }

    lat=new double[num_gridpoints];                     // latitude of every location
    lon=new double[num_gridpoints];                     // longitude of every location
    for(i=0;i<num_latpoints;i++){
      for(j=0;j<num_lonpoints;j++){
	lat[i*num_lonpoints+j]=lat_grid[i];
	lon[i*num_lonpoints+j]=lon_grid[j];
      }
    }    
  }
  fill_adj();        // finds adjacency info between all pairs of points, based on range variable

  //  permute();     // uses genqmd fortran function to permute points for maximal sparseness of Cholesky factor
  mat_pp<int> tmpperm(1,num_gridpoints);
  mat_pp<int> tmpinvp(1,num_gridpoints);
  tmpperm.read(perm_fn,BY_ROW,RAGGED);    // reads in the perm
  tmpinvp.read(invp_fn,BY_ROW,RAGGED);    // reads in the invp
  tmpperm.getvec(0,ROW,&perm);
  tmpinvp.getvec(0,ROW,&invp); 
  u=new mat_pp<double>(num_gridpoints,num_gridpoints);
  v=new mat_pp<double>(num_gridpoints,num_gridpoints);
  for(i=0;i<num_gridpoints;i++){
    for(int k=0;k<num_gridpoints;k++){      
      euc_distance(i,k,&(u->vals[i][k]),&(v->vals[i][k]),NO_PERMUTE);
    }
  }


  if(num_kernels){
    calc_locn_kernel_angdists(PERMUTE);
  }

  
}

locn_info::~locn_info(void){
  int i;
  delete [] adjncy;
  delete [] xadj;
  delete [] num_neighbors;
  for(i=0;i<num_gridpoints;i++){
    delete [] neighbor_ptr[i];
  }
  delete neighbor_ptr;
  delete [] lat;
  delete [] lon;
  if(format==GRID){
    delete [] lat_grid;
    delete [] lon_grid;
  }
  delete [] perm;
  delete [] invp;
  if(num_kernels){
    delete [] kernel_lat;
    delete [] kernel_lon;
    delete locn_kernel_dists;
  }
}

void locn_info::calc_locn_kernel_angdists(int permute){
  /* calcs angular distances in latitude and longitude between locations and centers of basis kernels */
  int i,k,permlocn;
  double diff;

  locn_kernel_dists=new mat_pp<double>(num_gridpoints,num_kernels);
  for(i=0;i<num_gridpoints;i++){
    if(permute==PERMUTE){   // lat/lon info not permuted, but cov_mat uses permuted order, so need to deal with this here
      permlocn=perm[i]-PERMSTART1;
    }
    else{
      permlocn=i;
    }
    for(k=0;k<num_kernels;k++){      
      diff=fabs(lon[permlocn]-kernel_lon[k]);
      
      locn_kernel_dists->vals[i][k]=acos(sin(lat[permlocn]*D2R)*sin(kernel_lat[k]*D2R)+cos(lat[permlocn]*D2R)*cos(kernel_lat[k]*D2R)*cos(diff*D2R));
    }
  }
}


void locn_info::euc_distance(int x,int y,double* euclid_u,double* euclid_v,int permute){
  /* calcs Euclidean u and v distances between locations x and y, x and y need to be unpermuted */
  double mid_lat_out; 
  double mid_lon_out;
  int permx,permy;
  if(permute==PERMUTE){                     
    permx=perm[x]-PERMSTART1;
    permy=perm[y]-PERMSTART1;
  }
  else{                      // NO_PERMUTE
    permx=x;
    permy=y;
  }

  // this code change to doing the u and v distance calcs before anything else was brought about by the incorrect u and v calcs being done with certain values of nu when the distance calcs are done when the correlation matrices are first filled; I'm still not sure what's causing the problem - it seems to happen on both h1 and h8; a good illustration is locations 90/210 (89/209) which are locations 2/133 (3/134) after permutation - the 2nd call to this function returns ~-1.02 as the distance when the real distance is 2.77 Mm

  if(permute==PERMUTE){
  *euclid_u=u->vals[permx][permy];
  *euclid_v=v->vals[permx][permy];
  }
  else{  // getting initial u and v values
    midpoint(lat[permx],lon[permx],lat[permy],lon[permy],&mid_lat_out,&mid_lon_out);   // finds lat/lon midpoint
    project(mid_lat_out,mid_lon_out,lat[permx],lon[permx],lat[permy],lon[permy],euclid_u,euclid_v,MM);   // finds Euclidean u and v distances between the two locations based on latitude/longitude and midpoint
    /*
      if(!zero(*euclid_u-u->vals[permx][permy]) || !zero(*euclid_v-v->vals[permx][permy])){
      cout << "WARNING: " <<   x << " " << y << " " << *euclid_u << " " << *euclid_v << " " << u->vals[permx][permy] << " " << v->vals[permx][permy] << " " << lat[permx] << " " << lat[permy] << " "<< lon[permx]<<  " "<< lon[permy] << endl;
    }
    if(printout==1){
      cout << x << " " << y << " " << *euclid_u << " " << *euclid_v << " " << lat[permx] << " " << lat[permy] << " "<< lon[permx]<<  " "<< lon[permy] << endl;
    }
    */
  }
}


double locn_info::ang_distance(int x,int y,int units,int permute){
  /* finds distance between locations x and y */

  double diff,dist;
  int permx,permy;
  if(permute==PERMUTE){                     
    permx=perm[x]-1;
    permy=perm[y]-1;
  }
  else{                      // NO_PERMUTE
    permx=x;
    permy=y;
  }
  diff=fabs(lon[permx]-lon[permy]);
  dist=acos(sin(lat[permx]*D2R)*sin(lat[permy]*D2R)+cos(lat[permx]*D2R)*cos(lat[permy]*D2R)*cos(diff*D2R));
  switch (units)
    {
    case RADIANS:
      return(dist);
    case DEGREES:
      return(dist*R2D);
    case MM:
      return(dist*EQ_RADIUS*M2MM);
    case KM:
      return(dist*EQ_RADIUS*M2KM);
    default:
      cerr << "Cannot calculate distance in units: " << units << endl;
      return(ERROR);
    }
}

void locn_info::fill_adj(void){
  /* determines pairs of locations for which correlations not assumed zero based on range variable, 
     fills adjacency structures based on this -- adjacency structures use indices 1,...,N 
     (Fortran style) to indicate which locations are neighbors of which locations
     but neighbor_ptr uses 0,...,N-1 

     This information is needed for smbfct symbolic factorization in cov_mat.C */

  int i, j, newi, tempnum;
  neighbor_ptr=new (int*)[num_gridpoints];                  // stores indices neighbors of each location
  int *tmp_neighbor_ptr=new int[num_gridpoints];            
  num_neighbors=new int[num_gridpoints];                    // stores number of neighbors each location has
  
  // not sure about +1 below -it wasn't in datstatsig but is in regr12
  int *tmp_adjncy=new int[num_gridpoints*num_gridpoints+1];              // adjacency vector in George & Liu form - Fortran style indexing - temporary since not sure how long it should be
  xadj=new int[num_gridpoints+1];                           // for each location, pointer to start of neighbors in tmp_adjncy/adjncy - Fortran style indexing
  
  adjncy_len=0;
  for(i=0;i<num_gridpoints;i++){
    newi=1;
    tempnum=0;
    for(j=0;j<num_gridpoints;j++){  
      if(i!=j && ang_distance(i,j,RADIANS,NO_PERMUTE)<range){
	tmp_adjncy[adjncy_len]=j+1;
	if(newi){
	  xadj[i]=adjncy_len+1;
	  newi=0;
	}
	adjncy_len++;
	if(i<j){
	  tmp_neighbor_ptr[tempnum]=j;
	}
	tempnum++;
      }
    }
    num_neighbors[i]=tempnum;
    neighbor_ptr[i]=new int[tempnum];
    for(j=0;j<tempnum;j++){
      neighbor_ptr[i][j]=tmp_neighbor_ptr[j];
    }
  }
  
  adjncy_len++;                                  // this leaves one empty space
  xadj[num_gridpoints]=adjncy_len;
  adjncy=new int[adjncy_len];                    // make vector to be exactly the right length
  for(i=0;i<adjncy_len;i++){
    adjncy[i]=tmp_adjncy[i];
  }
  delete [] tmp_adjncy;
  delete [] tmp_neighbor_ptr;

  cout << "number of points: " << adjncy_len << endl;
}

// not currently in use, since I'm not using sparse covariance matrix
void locn_info::permute(void){
  /* computes minimum degree ordering so that Choleski will be sparse if original cov matrix is sparse 
     i.e., permutes locations to increase sparseness of Cholesky */
  int i;
  int *marker; int* qsize; int* deg; int* qlink; int*nbrhd; int*rchset; 
  int nofsub;
  int *temp_adjncy;
  
  perm=new int[num_gridpoints];                       // stores permutation that achieves sparseness
  invp=new int[num_gridpoints];
  marker=new int[num_gridpoints];
  qsize=new int[num_gridpoints];
  qlink=new int[num_gridpoints];
  nbrhd=new int[num_gridpoints];
  rchset=new int[num_gridpoints];
  deg=new int[num_gridpoints];
  temp_adjncy=new int[adjncy_len];
  copy(adjncy_len,adjncy,temp_adjncy);      // genqmd destroys adjncy vector
  genqmd_(&num_gridpoints,xadj,temp_adjncy,perm,invp,deg,marker,rchset,nbrhd,qsize,qlink,&nofsub);  // Fortran function from George & Liu
  
  cout << "Permutation:";
  for(i=0;i<num_gridpoints;i++){
    cout << " " << perm[i];
  }
  cout << endl;
  
  delete [] marker;
  delete [] qsize;
  delete [] qlink;
  delete [] nbrhd;
  delete [] rchset;
  delete [] deg;
  delete [] temp_adjncy;
  
}
















